import inspect
import os
from typing import Any, Callable, Dict, Iterable, List

import dotenv


def infinite_iterator(iterator: Iterable) -> Any:
    """Infinite iterator

    :param iterator: iterator
    :type iterator: iterator
    :yield: item from iterator
    :rtype: Any
    """
    while True:
        for item in iterator:
            yield item


def property_with_cache(func: Callable) -> Callable:
    """Property decorator to cache the result of a property. The result is cached in the attribute of name "_{func.__name__}".

    :param func: function to decorate
    :type func: function
    :return: decorated function
    :rtype: function
    """

    @property
    def decorated_func(*args, **kwargs):
        raise RuntimeError("Deprecated. use functools.cached_property instead.")

    return decorated_func


def to_dict(input_object, exclude_prefixes: List[str] = ["_", "__"]) -> Dict[str, Any]:
    """Transform object to dictionary. Note that all the attributes that starts with "__" or callable are excluded.

    :return: Dictionary that contains all the attributes of the object
    :rtype: Dict
    """
    return {
        key: getattr(input_object, key)
        for key in dir(input_object)
        if all([not key.startswith(prefix) for prefix in exclude_prefixes])
        and not callable(getattr(input_object, key))
    }


def load_dotenv(stack_depth: int = 1):
    """Load dotenv from the order of PYTHONPATH, working directory, and the caller file path.

    :param stack_depth: stack_depth of this function from the target caller, defaults to 1
    :type stack_depth: int, optional
    """
    # Get possible paths
    python_paths = (
        os.environ["PYTHONPATH"].split(":") if "PYTHONPATH" in os.environ else []
    )
    working_dir_path = os.getcwd()
    caller_file_path = os.path.dirname(
        os.path.abspath(inspect.stack()[stack_depth].filename)
    )
    possible_paths = python_paths + [working_dir_path, caller_file_path]

    # Find and load dotenv
    for path in possible_paths:
        dotenv.load_dotenv(dotenv_path=os.path.join(path, ".env"))
        if "SLACK_ACCESS_TOKEN" in os.environ:
            break



import contextlib
import logging
import os

import attrs
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

import hkkang_utils.misc as misc_utils
import hkkang_utils.socket as socket_utils

# Load environment variables
misc_utils.load_dotenv(stack_depth=2)
# Get default access token
DEFAULT_ACCESS_TOKEN = (
    os.environ["SLACK_ACCESS_TOKEN"] if "SLACK_ACCESS_TOKEN" in os.environ else None
)

logger = logging.getLogger("SlackMessenger")


@attrs.define
class SlackMessenger:
    """Note that the default token is set by the environment variable SLACK_ACCESS_TOKEN."""

    channel: str = attrs.field()
    token: str = attrs.field(default=attrs.Factory(lambda: DEFAULT_ACCESS_TOKEN))
    append_src_info: bool = attrs.field(default=True)

    def __attrs_post_init__(self):
        if self.token is None:
            raise ValueError(
                "Please set token or set SLACK_ACCESS_TOKEN environment variable."
            )

    def send(self, text: str) -> None:
        return send_message(
            token=self.token,
            channel=self.channel,
            text=text,
            append_src_info=self.append_src_info,
        )


def send_message(
    channel: str,
    text: str,
    token: str = DEFAULT_ACCESS_TOKEN,
    append_src_info: bool = True,
) -> None:
    """Please follow the tutorial to get bot OAuthToken and setup the bot permissions.
    https://github.com/slackapi/python-slack-sdk/tree/main/tutorial
    """
    # Check if token is provided
    if token is None:
        raise ValueError(
            "Please set token or set SLACK_ACCESS_TOKEN environment variable."
        )
    # Create client
    client = WebClient(token=token)

    # Build message
    if append_src_info:
        ip = socket_utils.get_local_ip()
        host_name = socket_utils.get_host_name()
        text_with_prefix = f"Message from {host_name}({ip}):\n{text}"

    # Send message
    try:
        response = client.chat_postMessage(channel=channel, text=text_with_prefix)
        assert response["message"]["text"] == text_with_prefix
        logger.info(f"Sending message to channel {channel}: {text}")
    except SlackApiError as e:
        # You will get a SlackApiError if "ok" is False
        assert e.response["ok"] is False
        assert e.response["error"], "channel_not_found"
        logger.info(f"Got an error: {e.response['error']}")


@contextlib.contextmanager
def notification(
    channel: str,
    success_msg: str = None,
    error_msg: str = None,
    token: str = DEFAULT_ACCESS_TOKEN,
):
    slack_messenger = SlackMessenger(channel=channel, token=token)
    try:
        yield slack_messenger
        if success_msg is not None:
            slack_messenger.send(success_msg)
    except Exception as e:
        if error_msg is not None:
            slack_messenger.send(f"{error_msg} ({e.__class__.__name__}: {e})")
        raise e


@contextlib.contextmanager
def slack_notification(*args, **kwargs):
    raise NotImplementedError("Please use notification instead of slack_notification")



import csv
import inspect
import ujson
import os
import pathlib
import pickle
from typing import Any, Callable, Dict, List, Optional, Tuple

import omegaconf
import tqdm
import yaml


def get_files_in_directory(
    dir_path: str, filter_func: Optional[callable] = None, return_with_dir=False
) -> list:
    """Get paths of files in a directory
    :param dir_path: path of directory that you want to get files from
    :type dir_path: str
    :param filter_func: function that returns True if the file name is valid
    :type filter_func: callable
    :return: list of file paths in the directory which are valid
    :rtype: list
    """
    file_names = [
        f
        for f in os.listdir(dir_path)
        if os.path.isfile(os.path.join(dir_path, f))
        and (filter_func is None or filter_func(f))
    ]
    if return_with_dir:
        return [os.path.join(dir_path, f) for f in file_names]
    return file_names


def get_files_in_all_sub_directories(
    root_dir_path: str, filter_func: Optional[callable] = None
) -> list:
    """Get paths of files in all sub directories
    :return: list of file paths in all sub directories which are valid
    :rtype: list
    """
    return [
        os.path.join(dp, f)
        for dp, dn, filenames in os.walk(root_dir_path)
        for f in filenames
        if (filter_func is None or filter_func(f))
    ]


def create_directory(dir_path: str):
    """Creates all directories of the given path (if not exists)

    :param dir_path: directory path
    :type dir_path: str
    """
    return pathlib.Path(dir_path).mkdir(parents=True, exist_ok=True)


def split_path_into_dir_and_file_name(file_path: str) -> Tuple[str, str]:
    """Split a file path into directory path and file name

    :param file_path: file path
    :type file_path: str
    :return: directory path and file name
    :rtype: Tuple[str, str]
    """
    return os.path.dirname(file_path), os.path.basename(file_path)


def path_from_current_file(relative_path: str) -> str:
    """convert relative path to absolute path with respect to the caller's file path"""
    file_path_of_caller = inspect.stack()[1].filename
    dir_path_of_caller = os.path.dirname(file_path_of_caller)
    full_path = os.path.join(dir_path_of_caller, relative_path)
    return os.path.normpath(full_path)


# Related to json files
def read_json_file(
    file_path: str,
    auto_detect_extension: bool = False,
    encoding: Optional[str] = None,
) -> Dict:
    """Read a json file

    :param file_path: json file path
    :type file_path: str
    :return: json data
    :rtype: dict
    """
    if auto_detect_extension and file_path.endswith(".jsonl"):
        return read_jsonl_file(file_path, encoding=encoding)
    else:
        with open(file_path, "r", encoding=encoding) as f:
            return ujson.load(f)


def read_jsonl_file(file_path: str, encoding: Optional[str] = None) -> Dict:
    with open(file_path, "r", encoding=encoding) as f:
        return [ujson.loads(line) for line in f.readlines()]


def write_json_file(
    dict_object: Dict,
    file_path: str,
    indent: int = 4,
    auto_detect_extension: bool = False,
    encoding: Optional[str] = None,
) -> None:
    if auto_detect_extension and file_path.endswith(".jsonl"):
        return write_jsonl_file(
            dict_object=dict_object, file_path=file_path, encoding=encoding
        )
    else:
        with open(file_path, "w", encoding=encoding) as f:
            ujson.dump(dict_object, f, indent=indent)


def write_jsonl_file(
    dict_object: Dict,
    file_path: str,
    encoding: Optional[str] = None,
):
    with open(file_path, "w", encoding=encoding) as f:
        for line in dict_object:
            f.write(f"{ujson.dumps(line)}\n")


# Related to yaml files
def read_yaml_file(file_path: str) -> Dict:
    """Read a yaml file

    :param file_path: yaml file path
    :type file_path: str
    :return: yaml data
    :rtype: dict
    """
    with open(file_path, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return None


def write_yaml_file(dict_object: dict, file_path: str) -> None:
    with open(file_path, "w") as yaml_file:
        yaml.dump(dict_object, yaml_file, default_flow_style=False)


# Related to pickle files
def write_pickle_file(object_to_save, file_path: str) -> None:
    """Write a pickle file

    :param object_to_save: object to save
    :type object_to_save: any
    :param file_path: pickle file path
    :type file_path: str
    """
    with open(file_path, "wb") as f:
        pickle.dump(object_to_save, f)


def read_pickle_file(file_path: str) -> any:
    """Read a pickle file

    :param file_path: pickle file path
    :type file_path: str
    :return: object
    :rtype: any
    """
    with open(file_path, "rb") as f:
        return pickle.load(f)


# Related to config files


def read_config_file(file_path: str) -> omegaconf.OmegaConf:
    """Load a config file

    :param file_path: config file path
    :type file_path: str
    :return: config data
    :rtype: omegaconf.OmegaConf
    """
    return omegaconf.OmegaConf.load(file_path)


def load_config_file(file_path: str) -> omegaconf.OmegaConf:
    """Load a config file

    :param file_path: config file path
    :type file_path: str
    :return: config data
    :rtype: omegaconf.OmegaConf
    """
    raise RuntimeError("load_config_file is deprecated. Use read_config_file instead.")


def write_config_file(config: omegaconf.OmegaConf, file_path: str) -> None:
    """Write a config file

    :param config: config data
    :type config: omegaconf.OmegaConf
    :param file_path: config file path
    :type file_path: str
    """
    omegaconf.OmegaConf.save(config, file_path)


# Related to csv file
def read_csv_file(
    file_path: str,
    delimiter: str = ",",
    quotechar: str = '"',
    show_progress: bool = False,
    process_row_func: Callable = None,
) -> List[Dict[str, Any]]:
    dict_list = []
    with open(file_path, "r") as f:
        tsv_file_reader = csv.reader(f, delimiter=delimiter, quotechar=quotechar)
        # Get headers
        header = next(tsv_file_reader)
        # Get values
        file_iterator = tqdm.tqdm(tsv_file_reader) if show_progress else tsv_file_reader
        for row in file_iterator:
            row = process_row_func(row) if process_row_func else row
            dict_list.append(dict(zip(header, row)))
    return dict_list


def write_csv_file(
    list_of_dict: List[Dict[str, Any]], file_path: str, delimiter: str = ","
) -> None:
    with open(file_path, "w") as f:
        tsv_file_writer = csv.writer(f, delimiter=delimiter)
        tsv_file_writer.writerow(list_of_dict[0].keys())
        for dict_item in list_of_dict:
            tsv_file_writer.writerow(dict_item.values())



import time as time_module
import unittest
from typing import List, Tuple

import src.hkkang_utils.time as time_utils

TEST_TIME = 2


class Test_time_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_time_utils, self).__init__(*args, **kwargs)

    def test_timer_measure(self):
        timer = time_utils.Timer()
        timer.start()
        time_module.sleep(TEST_TIME)
        timer.stop()
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."

    def test_timer_measure_using_with(self):
        timer = time_utils.Timer()
        with timer.measure():
            time_module.sleep(TEST_TIME)
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."

    def test_timer_singleton_by_name1(self):
        def test_func():
            timer = time_utils.Timer()
            with timer.measure():
                time_module.sleep(TEST_TIME)

        test_func()
        timer = time_utils.Timer(
            class_name="Test_time_utils",
            func_name="test_timer_singleton_by_name1.test_func",
        )
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."

    def test_timer_singleton_by_name2(self):
        timer = time_utils.Timer()
        with timer.measure():
            time_module.sleep(TEST_TIME)

        timer = time_utils.Timer(
            class_name="Test_time_utils",
            func_name="test_timer_singleton_by_name2",
        )
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."

    def test_timer_decorator(self):
        @time_utils.measure_time
        def test_func():
            time_module.sleep(TEST_TIME)

        test_func()
        timer = time_utils.Timer(
            class_name="Test_time_utils", func_name="test_timer_decorator.test_func"
        )
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."

    def test_timer_measure_total_elapsed_time(self):
        timer = time_utils.Timer()
        with timer.measure():
            time_module.sleep(TEST_TIME)
        with timer.measure():
            time_module.sleep(TEST_TIME)
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."
        self.assertTrue(
            timer.total_elapsed_time > TEST_TIME * 2
            and timer.total_elapsed_time < (TEST_TIME * 2 + 1),
            f"Timer is not working properly: {timer.total_elapsed_time} sec is measured.",
        )
        timer.show_total_elapsed_time()

    def test_timer_measure_avg_elapsed_time(self):
        timer = time_utils.Timer()
        with timer.measure():
            time_module.sleep(TEST_TIME)
        with timer.measure():
            time_module.sleep(TEST_TIME)
        self.assertTrue(
            timer.elapsed_time > TEST_TIME and timer.elapsed_time < TEST_TIME + 1
        ), f"Timer is not working properly: {timer.elapsed_time} sec is measured."
        self.assertTrue(
            timer.avg_elapsed_time > TEST_TIME
            and timer.avg_elapsed_time < (TEST_TIME + 1),
            f"Timer is not working properly: {timer.avg_elapsed_time} sec is measured.",
        )
        timer.show_avg_elapsed_time()


class Test_time_utils2(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_time_utils2, self).__init__(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        timer1 = time_utils.Timer(func_name="test_timer_summary1")
        timer2 = time_utils.Timer(func_name="test_timer_summary2")
        # Measure time
        with timer1.measure():
            time_module.sleep(TEST_TIME)
        with timer1.measure():
            time_module.sleep(TEST_TIME)
        with timer2.measure():
            time_module.sleep(TEST_TIME)
        with timer2.measure():
            time_module.sleep(TEST_TIME)

    def test_timer_summary_for_single_instance(self):
        timer = time_utils.Timer(func_name="test_timer_summary1")
        # Get statistics
        fname, call_cnt, avg_time, total_time = timer.summarize_measured_time()
        # Check
        self.assertEqual(fname, "test_timer_summary1")
        self.assertEqual(call_cnt, 2)
        self.assertGreater(avg_time, TEST_TIME)
        self.assertLess(avg_time, TEST_TIME + 1)
        self.assertGreater(total_time, TEST_TIME * 2)
        self.assertLess(total_time, TEST_TIME * 2 + 1)

    def test_timer_summary_for_all_instances(self):
        infos: List[
            Tuple[str, int, float, float]
        ] = time_utils.Timer.summarize_measured_times()
        fnames = list(map(lambda k: k[0], infos))
        self.assertGreaterEqual(len(infos), 2)
        self.assertIn("test_timer_summary1", fnames)
        test_timer_summary1_idx = fnames.index("test_timer_summary1")
        self.assertEqual(infos[test_timer_summary1_idx][1], 2)
        self.assertGreater(infos[test_timer_summary1_idx][2], TEST_TIME)
        self.assertLess(infos[test_timer_summary1_idx][2], TEST_TIME + 1)
        self.assertGreater(infos[test_timer_summary1_idx][3], TEST_TIME * 2)
        self.assertLess(infos[test_timer_summary1_idx][3], TEST_TIME * 2 + 1)
        self.assertIn("test_timer_summary2", fnames)
        test_timer_summary2_idx = fnames.index("test_timer_summary2")
        self.assertEqual(infos[test_timer_summary2_idx][1], 2)
        self.assertGreater(infos[test_timer_summary2_idx][2], TEST_TIME)
        self.assertLess(infos[test_timer_summary2_idx][2], TEST_TIME + 1)
        self.assertGreater(infos[test_timer_summary2_idx][3], TEST_TIME * 2)
        self.assertLess(infos[test_timer_summary2_idx][3], TEST_TIME * 2 + 1)


if __name__ == "__main__":
    unittest.main()
